function fh = pbar(n, fh)
% # parfor-compatible progress bar
% ## Usage
% * `fh = pbar(N)`
% * `pbar(n, fh)`
%
% ## Description
% A progress bar for monitoring the progress of, for example, Monte Carlo
% simulations, iterative algorithms, etc. The progress bar is parfor-
% compatible and based on the comment by 'Philip' in [1] but uses a 
% completely random file name, in case multiple instances of Matlab are
% running in parallel.
% 
% ## Input
% * `N`: Number of iterations, to be used before and after the loop.
% * `fh`: File handle to an existing progress bar, to be used inside the
%   loop.
%
% ## Output
% * `fh`: File handle, returned after the first call to `pbar`.
% 
% ## Example
% fh = pbar(N);       % Initialize progress bar
% for n = 1:N
%     % Many time-consuming operations
%     % ...
%
%     pbar(n, fh);    % Update the progress
% end
% pbar(0, fh);        % Reset progress bar
% 
% ## References
% * [1] https://se.mathworks.com/matlabcentral/fileexchange/32101-progress-monitor--progress-bar--that-works-with-parfor
%
% ## Authors
% 2017-present -- Roland Hostettler

%{
% This file is free software: you can redistribute it and/or modify it 
% under the terms of the GNU General Public License as published by the 
% Free Software Foundation, either version 3 of the License, or (at your
% option) any later version.
% 
% This file is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
% FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
% more details.
% 
% You should have received a copy of the GNU General Public License along 
% with this file. If not, see <http://www.gnu.org/licenses/>.
%}

% TODO:
% * Make it compatible such that we can call it w/o file handle

    %% Sanity Checks and Defaults
    narginchk(1, 2);    
    width = 20;
    
    %% Progress Display
    if nargin == 1 && n > 0
        % Initializate a temporary file
        fh = tempname();
        f = fopen(fh, 'w');
        if f < 0
            error('Could not open temporary file %s.', fh);
        end
        
        % Write the data:
        %   - First word is the no. of total iterations
        %   - Second word is the no. of iteration so far
        fwrite(f, n, 'uint32');
        fwrite(f, 0, 'uint32');
        fclose(f);
        disp(['  0 % [>', repmat(' ', 1, width), ']']);
        
    elseif nargin == 2 && n == 0
        % Done; remove the temporary file and show 100 %
        delete(fh);
        delbar = repmat(char(8), 1, (width+11));
        bar = repmat('=', 1, width+1);
        disp([delbar, char(10), '100 % [', bar, ']']);
    else
        if ~exist(fh, 'file')
            error('Temporary file not found. Run pbar(N) before pbar(n, fh) to initialize.');
        end
        f = fopen(fh, 'r+');
        A = fread(f, 2, 'uint32');
        N = A(1);
        n = A(2) + 1;
        fseek(f, 4, 'bof');
        fwrite(f, n, 'uint32');
        fclose(f);

        percent = sprintf('%3.0f %% ', n/N*100);
        delbar = repmat(char(8), [1, (width+11)]);
        bar = repmat('=', [1, round(n/N*width)]);
        blanks = repmat(' ', [1, width - round(n/N*width)]);
        disp([delbar, char(10), percent, '[', bar, '>', blanks, ']']);
    end 
end
